# -*- coding: utf-8 -*-
import json
from abc import ABC

__all__ = (
    'Snowflake',
    'Designation'
)

class Snowflake(ABC):

    """
    Abstract base class depicting a Discord object that has an ID.
    """

    __slots__ = ('id')

    def __init__(self, **kwargs):
        self.id = kwargs.get('id')

    def __str__(self) -> str:
        return f'<{type(self).__name__} id={self.id}>'

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, other) -> bool:
        return self.id == other.id

class Designation(Snowflake):

    """
    An extention to the Snowflake abstract base class depicting a Discord object that has a name.
    """

    __slots__ = ('name')

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.name = kwargs.get('name')

    def __str__(self) -> str:
        return f'<{type(self).__name__} id={self.id}, name={self.name}>'



__title__ = 'restcord'
__author__ = 'Lethys'
__license__ = 'MIT'
__copyright__ = 'Copyright 2020 (c) Lethys'
__version__ = '0.0.5'

from .ban import Ban
from .channel import Channel
from .client import RestCord
from .emoji import Emoji
from .errors import (
    BadGateway,
    BadRequest,
    Forbidden,
    HTTPException,
    InternalServerError,
    NotFound,
    RateLimited
)
from .guild import Guild, GuildPreview
from .member import Member
from .message import Message
from .role import Role
from .user import User
from .voice import VoiceRegion



# -*- coding: utf-8 -*-
from .permissions import PermissionOverwrite
from .snowflake import Designation

__all__ = (
    'Channel'
)

class Channel(Designation):

    """
    Model depicting a Discord channel object.
    """

    __slots__ = ('guild_id', 'type', 'position', 'permission_overwrites', 'parent_id', 'last_message_id', 'last_pin_timestamp', 'topic', 'nsfw', 'permission_overwrites')

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.guild_id = kwargs.get('guild_id')
        self.type = kwargs.get('type')
        self.position = kwargs.get('position')
        self.permission_overwrites = kwargs.get('permission_overwrites')
        self.parent_id = kwargs.get('parent_id')
        self.last_message_id = kwargs.get('last_message_id')
        self.last_pin_timestamp = kwargs.get('last_pin_timestamp')
        self.topic = kwargs.get('topic')
        self.nsfw = kwargs.get('nsfw')
        self.permission_overwrites = [PermissionOverwrite(**p) for p in kwargs.get('permission_overwrites', [])]

    def __str__(self):
        return f'<{type(self).__name__} id={self.id}, name={self.name}, type={self.type}, position={self.position}>'

    @property
    def mention(self) -> str:
        """:class:`str`: The channel's mentionable string."""
        return f'<#{self.id}>'



# -*- coding: utf-8 -*-
import logging
from typing import Optional

from aiohttp import ClientSession

from .channel_client import ChannelClient
from .emoji_client import EmojiClient
from .guild_client import GuildClient
from .user_client import UserClient
from .voice_client import VoiceClient

__log__ = logging.getLogger(__name__)

__all__ = (
    'RestCord'
)

class RestCord:

    """
    Asynchronous Python clinet for communicating with Discord's API.

    Parameters
    ------------
    token: str
        Your application's token from: https://discord.com/developers/applications
    session: Optional[ClientSession]
        Optionally include your aiohttp session
    """

    __slots__ = ('channel_client', 'emoji_client', 'guild_client', 'user_client', 'voice_client')

    def __init__(self, token: str, loop=None, proxy=None, proxy_auth=None, session: Optional[ClientSession]=None) -> None:
        self.channel_client = ChannelClient(token=token, loop=loop, proxy=proxy, proxy_auth=proxy_auth, session=session)
        self.emoji_client = EmojiClient(token=token, loop=loop, proxy=proxy, proxy_auth=proxy_auth, session=session)
        self.guild_client = GuildClient(token=token, loop=loop, proxy=proxy, proxy_auth=proxy_auth, session=session)
        self.user_client = UserClient(token=token, loop=loop, proxy=proxy, proxy_auth=proxy_auth, session=session)
        self.voice_client = VoiceClient(token=token, loop=loop, proxy=proxy, proxy_auth=proxy_auth, session=session)

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.close()

    async def close(self):
        await self.channel_client.close()
        await self.emoji_client.close()
        await self.guild_client.close()
        await self.user_client.close()
        await self.voice_client.close()



# -*- coding: utf-8 -*-
from .snowflake import Designation
from .user import User

__all__ = (
    'Emoji'
)

class Emoji(Designation):

    """
    Model depicting a Discord emoji object.
    """

    __slots__ = ('roles', 'require_colons', 'managed', 'animated', 'available', 'user')

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.roles = kwargs.get('roles')
        self.require_colons = kwargs.get('require_colons')
        self.managed = kwargs.get('managed')
        self.animated = kwargs.get('animated')
        self.available = kwargs.get('available')

        user = kwargs.get('user')
        if user:
            self.user = User(**user)
        else:
            self.user = None

    @property
    def mention(self) -> str:
        """:class:`str`: The emoji's mentionable string."""
        if self.animated:
            return f'<a:{self.name}:{self.id}>'
        return f'<:{self.name}:{self.id}>'



# -*- coding: utf-8 -*-
import logging
from typing import List, Optional

from aiohttp import ClientSession

from .http import HTTPClient, Route
from .message import Message
from .user import User

__log__ = logging.getLogger(__name__)

__all__ = (
    'ChannelClient'
)

class ChannelClient(HTTPClient):

    def __init__(self, token: str, loop=None, proxy=None, proxy_auth=None, session: Optional[ClientSession]=None) -> None:
        super().__init__(token=token, loop=loop, proxy=proxy, proxy_auth=proxy_auth, session=session)

    async def get_message(self, channel_id: int, message_id: int) -> Message:
        """|coro|
        Get a channel's message.

        Returns
        ---------
        Optional[:class:`Message`]
            The Message or ``None`` if not found.

        API Documentation
        ----------
            https://discord.com/developers/docs/resources/channel#get-channel-message

        Parameters
        ----------
        channel_id: :class:`int`
            Discord's identifier for the channel.
        message_id: :class:`int`
            Discord's identifier for the message.
        """

        if not channel_id:
            raise ValueError("Argument cannot be None: channel_id")

        if not message_id:
            raise ValueError("Argument cannot be None: message_id")

        message = await self._request(Route('GET', f'/channels/{channel_id}/messages/{message_id}'))

        return Message(**message)

    async def get_messages(self, channel_id: int, around=None, before=None, after=None, limit=50) -> List[Message]:
        """|coro|
        Get a list of a channel's messages.

        Returns
        ---------
        List[:class:`Message`]:
            The list of Channels.

        API Documentation
        ----------
            https://discord.com/developers/docs/resources/channel#get-channel-messages

        Parameters
        ----------
        channel_id: :class:`int`
            Discord's identifier for the channel.
        around: Optional[:class:`int`]
	        Get messages around this message ID
        before: Optional[:class:`int`]
	        Get messages before this message ID
        after: Optional[:class:`int`]
	        Get messages after this message ID
        limit: Optional[:class:`int`]
	        Max number of messages to return (1-100).
            Defaults to 50.
        """

        if not channel_id:
            raise ValueError("Argument cannot be None: channel_id")

        params = {
            'limit': limit
        }

        if before is not None:
            params['before'] = before

        if after is not None:
            params['after'] = after

        if around is not None:
            params['around'] = around

        messages = await self._request(Route('GET', f'/channels/{channel_id}/messages'), params=params)

        return [Message(**message) for message in messages]

    async def add_reaction(self, channel_id: int, message_id: int, emoji: str):
        """|coro|
        Add a reaction to a message.

        API Documentation
        ----------
            https://discord.com/developers/docs/resources/channel#create-reaction

        Parameters
        ----------
        channel_id: :class:`int`
            Discord's identifier for the channel.
        message_id: :class:`int`
            Discord's identifier for the message.
        emoji: :class:`str`
            The URL encoded emoji
        """

        if not channel_id:
            raise ValueError("Argument cannot be None: channel_id")

        if not message_id:
            raise ValueError("Argument cannot be None: message_id")

        if not emoji:
            raise ValueError("Argument cannot be None: emoji")

        await self._request(Route('PUT', f'/channels/{channel_id}/messages/{message_id}/reactions/{emoji}/@me'))

    async def get_reactions(self, channel_id: int, message_id: int, emoji: str, before=None, after=None, limit=25) -> List[User]:
        """|coro|
        Get a list of users who have reacted to this message with the emoji.

        Returns
        ---------
        List[:class:`User`]:
            The list of Users who have added this reaction.

        API Documentation
        ----------
            https://discord.com/developers/docs/resources/channel#get-reactions

        Parameters
        ----------
        channel_id: :class:`int`
            Discord's identifier for the channel.
        message_id: :class:`int`
            Discord's identifier for the message_id.
        emoji: :class:`str`
            Discord's identifier for the channel.
        before: Optional[:class:`int`]
	        Get reactions before this user ID
        after: Optional[:class:`int`]
	        Get reactions after this user ID
        limit: Optional[:class:`int`]
	        Max number of users to return.
            Defaults to 25.
        """

        if not channel_id:
            raise ValueError("Argument cannot be None: channel_id")

        if not message_id:
            raise ValueError("Argument cannot be None: message_id")

        if not emoji:
            raise ValueError("Argument cannot be None: emoji")

        params = {
            'limit': limit
        }

        if before is not None:
            params['before'] = before

        if after is not None:
            params['after'] = after

        users = await self._request(Route('GET', f'/channels/{channel_id}/messages/{message_id}/reactions/{emoji}'), params=params)

        return [User(**user) for user in users]


# -*- coding: utf-8 -*-
import logging
from typing import List, Optional

from aiohttp import ClientSession

from .http import HTTPClient, Route
from .user import User

__log__ = logging.getLogger(__name__)

__all__ = (
    'UserClient'
)

class UserClient(HTTPClient):

    def __init__(self, token: str, loop=None, proxy=None, proxy_auth=None, session: Optional[ClientSession]=None) -> None:
        super().__init__(token=token, loop=loop, proxy=proxy, proxy_auth=proxy_auth, session=session)

    async def get_user(self, user_id: int) -> User:
        """|coro|
        Get a user.

        Returns
        ---------
        Optional[:class:`User`]
            The User or ``None`` if not found.

        API Documentation
        ----------
            https://discord.com/developers/docs/resources/user#get-user

        Parameters
        ----------
        user_id: :class:`int`
            Discord's identifier for the user.
        """

        if not user_id:
            raise ValueError("Argument cannot be None: user_id")

        user = await self._request(Route('GET', f'/users/{user_id}'))

        return User(**user)



# -*- coding: utf-8 -*-
import asyncio
import datetime
import json
import logging
import sys
from typing import Optional
from urllib.parse import quote as _uriquote

import aiohttp
from aiohttp import ClientSession

from . import __version__
from .errors import (
    BadGateway,
    BadRequest,
    Forbidden,
    HTTPException,
    InternalServerError,
    NotFound,
    RateLimited
)

__log__ = logging.getLogger(__name__)

__all__ = (
    'Route',
    'HTTPClient'
)

class Route:

    BASE = 'https://discord.com/api'

    def __init__(self, method, path):
        self.path = path
        self.method = method
        self.url = (self.BASE + self.path)

class HTTPClient:

    __slots__ = ('token', 'loop', 'proxy', 'proxy_auth', '__session', '__agent')

    def __init__(self, token: str, loop=None, proxy=None, proxy_auth=None, session: Optional[ClientSession]=None) -> None:
        self.token = token
        self.loop = asyncio.get_event_loop() if loop is None else loop
        self.proxy = proxy
        self.proxy_auth = proxy_auth
        self.__session = session
        self.__agent = f'RestCord.py (https://github.com/Yandawl/restcord.py {__version__}) Python/{sys.version_info[0]}.{sys.version_info[1]} aiohttp/{aiohttp.__version__}'

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.close()

    @property
    def session(self) -> ClientSession:
        """:class:`ClientSession`: The aiohttp ClientSession."""
        if self.__session is None or self.__session.closed:
            self.__session = ClientSession()
        return self.__session

    async def close(self):
        if self.__session:
            await self.__session.close()

    async def _request(self, route: Route, **kwargs):
        method = route.method
        url = route.url

        kwargs['headers'] = {
            'User-Agent': self.__agent,
            'X-Ratelimit-Precision': 'millisecond',
            'Authorization': f'Bot {self.token}'
        }

        if 'json' in kwargs:
            kwargs['headers']['Content-Type'] = 'application/json'
            kwargs['data'] = self.__to_json(kwargs.pop('json'))

        if self.proxy is not None:
            kwargs['proxy'] = self.proxy

        if self.proxy_auth is not None:
            kwargs['proxy_auth'] = self.proxy_auth

        async with self.session.request(method, url, **kwargs) as r:
            __log__.debug(f'{method} {url} with {kwargs.get("data")} has returned {r.status}')

            data = await self.__get_data(r)

            remaining = r.headers.get('X-Ratelimit-Remaining')
            if remaining == '0' and r.status != 429:
                __log__.debug(f'A rate limit bucket has been exhausted (retry: {self.__parse_ratelimit_header(r)}).')

            if 300 > r.status >= 200:
                __log__.debug(f'{method} {url} has received {data}')
                return data

            if r.status == 429:
                raise RateLimited(r, data)

            if r.status == 400:
                raise BadRequest(r, data)

            if r.status == 403:
                raise Forbidden(r, data)

            if r.status == 404:
                raise NotFound(r, data)

            if r.status == 500:
                raise InternalServerError(r, data)

            if r.status == 502:
                raise BadGateway(r, data)

            raise HTTPException(r, data)

    async def __get_data(self, response):
        text = await response.text(encoding='utf-8')
        try:
            if response.headers['content-type'] == 'application/json':
                return json.loads(text)
        except KeyError:
            pass

        return text

    def __parse_ratelimit_header(self, request, *, use_clock=False):
        reset_after = request.headers.get('X-Ratelimit-Reset-After')
        if use_clock or not reset_after:
            utc = datetime.timezone.utc
            now = datetime.datetime.now(utc)
            reset = datetime.datetime.fromtimestamp(float(request.headers['X-Ratelimit-Reset']), utc)
            return (reset - now).total_seconds()
        else:
            return float(reset_after)
    
    def __to_json(self, obj):
        return json.dumps(obj, separators=(',', ':'), ensure_ascii=True)



# -*- coding: utf-8 -*-
from .snowflake import Designation

__all__ = (
    'Role'
)

class Role(Designation):

    """
    Model depicting a Discord role object.
    """

    __slots__ = ('permissions', 'position', 'color', 'hoist', 'managed', 'mentionable', 'permissions_new')

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.permissions = kwargs.get('permissions')
        self.position = kwargs.get('position')
        self.color = kwargs.get('color')
        self.hoist = kwargs.get('hoist')
        self.managed = kwargs.get('managed')
        self.mentionable = kwargs.get('mentionable')
        self.permissions_new = kwargs.get('permissions_new')

    @property
    def mention(self) -> str:
        """:class:`str`: The role's mentionable string."""
        return f'<@&{self.id}>'



# -*- coding: utf-8 -*-
from restcord import utils

from .user import User

__all__ = (
    'Member'
)

class Member(User):

    """
    Model depicting a Discord member object.
    A member is an extention of :class:`User` where user is a member of a guild.
    """

    __slots__ = ('nick', 'premium_since', 'mute', 'deaf', 'joined_at')

    def __init__(self, **kwargs):
        user = kwargs.get('user')

        self.id = user.get('id')
        self.name = user.get('username')
        self.discriminator = user.get('discriminator')
        self.avatar = user.get('avatar')
        self.nick = kwargs.get('nick')
        self.premium_since = utils.parse_time(kwargs.get('premium_since'))
        self.mute = kwargs.get('mute')
        self.deaf = kwargs.get('deaf')
        self.joined_at = utils.parse_time(kwargs.get('joined_at'))

    def __str__(self):
        return f'<{type(self).__name__} id={self.id}, name={self.name}, discriminator={self.discriminator}, nick={self.nick}, joined_at={self.joined_at}>'



# -*- coding: utf-8 -*-
from restcord import utils

from .snowflake import Snowflake
from .user import User

__all__ = (
    'Message'
)

class Message(Snowflake):

    __slots__ = ('channel_id', 'type', 'content', 'author', 'pinned', 'tts', 'mention_everyone', 'timestamp', 'edited_timestamp')

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.channel_id = kwargs.get('channel_id')
        self.type = kwargs.get('type')
        self.content = kwargs.get('content')

        author = kwargs.get('author')
        if author:
            self.author = User(**author)
        else:
            self.author = None

        self.pinned = kwargs.get('pinned')
        self.tts = kwargs.get('tts')
        self.mention_everyone = kwargs.get('mention_everyone')
        
        self.timestamp = utils.parse_time(kwargs.get('timestamp'))

        edited_timestamp = kwargs.get('edited_timestamp')
        if edited_timestamp:
            self.edited_timestamp = utils.parse_time(edited_timestamp)
        else:
            self.edited_timestamp = None

    def __str__(self) -> str:
        return f'<{type(self).__name__} id={self.id}, channel_id={self.channel_id}, timestamp={self.timestamp}>'



# -*- coding: utf-8 -*-
import logging
from typing import List, Optional

from aiohttp import ClientSession

from .ban import Ban
from .channel import Channel
from .guild import Guild, GuildPreview
from .http import HTTPClient, Route
from .member import Member
from .role import Role

__log__ = logging.getLogger(__name__)

__all__ = (
    'GuildClient'
)

class GuildClient(HTTPClient):

    def __init__(self, token: str, loop=None, proxy=None, proxy_auth=None, session: Optional[ClientSession]=None) -> None:
        super().__init__(token=token, loop=loop, proxy=proxy, proxy_auth=proxy_auth, session=session)

    async def get_guild(self, guild_id: int, with_counts=False) -> Guild:
        """|coro|
        Get a guild.

        Returns
        ---------
        Optional[:class:`Guild`]
            The Guild or ``None`` if not found.

        API Documentation
        ----------
            https://discord.com/developers/docs/resources/guild#get-guild

        Parameters
        ----------
        guild_id: :class:`int`
            Discord's identifier for the guild.
        with_counts: :class:`bool`
            Whether to include approximate member counts.
            Defaults to ``False``.
        """

        if not guild_id:
            raise ValueError("Argument cannot be None: guild_id")

        params = {
            'with_counts': int(with_counts)
        }

        guild = await self._request(Route("GET", f'/guilds/{guild_id}'), params=params)

        return Guild(**guild)

    async def get_guild_preview(self, guild_id: int) -> GuildPreview:
        """|coro|
        Get a guild preview.

        Returns
        ---------
        Optional[:class:`GuildPreview`]
            The GuildPreview or ``None`` if not found.

        API Documentation
        ----------
            https://discord.com/developers/docs/resources/guild#guild-preview-object

        Parameters
        ----------
        guild_id: :class:`int`
            Discord's identifier for the guild.
        """

        if not guild_id:
            raise ValueError("Argument cannot be None: guild_id")

        guild = await self._request(Route("GET", f'/guilds/{guild_id}/preview'))

        return GuildPreview(**guild)

    async def get_member(self, guild_id: int, member_id: int) -> Member:
        """|coro|
        Get a guild's member.

        Returns
        ---------
        Optional[:class:`Member`]
            The Member or ``None`` if not found.

        API Documentation
        ----------
            https://discord.com/developers/docs/resources/guild#get-guild-member

        Parameters
        ----------
        guild_id: :class:`int`
            Discord's identifier for the guild.
        member_id: :class:`int`
            Discord's identifier for the member.
        """

        if not guild_id:
            raise ValueError("Argument cannot be None: guild_id")

        if not member_id:
            raise ValueError("Argument cannot be None: member_id")

        member = await self._request(Route('GET', f'/guilds/{guild_id}/members/{member_id}'))

        return Member(**member)

    async def get_members(self, guild_id: int, limit: int=1, after_id: int=0) -> List[Member]:
        """|coro|
        Get a list of a guild's members.

        Returns
        ---------
        List[:class:`Member`]:
            The list of Members.

        API Documentation
        ----------
            https://discord.com/developers/docs/resources/guild#list-guild-members

        Parameters
        ----------
        guild_id: :class:`int`
            Discord's identifier for the guild.
        limit: Optional[:class:`int`]
            Limit the amount of members returned.
            Defaults to ``1``.
        after_id: Optional[:class:`int`]
            Only get members with an id greater than after_id.
            Defaults to ``0``.
        """

        if not guild_id:
            raise ValueError("Argument cannot be None: guild_id")

        params = {
            'limit': limit,
            'after': after_id
        }

        members = await self._request(Route('GET', f'/guilds/{guild_id}/members'), params=params)

        return [Member(**member) for member in members]

    async def get_channel(self, channel_id: int) -> Channel:
        """|coro|
        Get a guild's channels.

        Returns
        ---------
        Optional[:class:`Channel`]
            The Channel or ``None`` if not found.

        API Documentation
        ----------
            https://discord.com/developers/docs/resources/channel#get-channel

        Parameters
        ----------
        channel_id: :class:`int`
            Discord's identifier for the channel.
        """

        if not channel_id:
            raise ValueError("Argument cannot be None: channel_id")

        channel = await self._request(Route('GET', f'/channels/{channel_id}'))

        return Channel(**channel)

    async def get_channels(self, guild_id: int) -> List[Channel]:
        """|coro|
        Get a list of a guild's channels.

        Returns
        ---------
        List[:class:`Channel`]:
            The list of Channels.

        API Documentation
        ----------
            https://discord.com/developers/docs/resources/guild#get-guild-channels

        Parameters
        ----------
        guild_id: :class:`int`
            Discord's identifier for the guild.
        """

        if not guild_id:
            raise ValueError("Argument cannot be None: guild_id")

        channels = await self._request(Route('GET', f'/guilds/{guild_id}/channels'))

        return [Channel(**channel) for channel in channels]

    async def get_roles(self, guild_id: int) -> List[Role]:
        """|coro|
        Get a list of a guild's roles.

        Returns
        ---------
        List[:class:`Role`]:
            The list of Roles.

        API Documentation
        ----------
            https://discord.com/developers/docs/resources/guild#get-guild-roles

        Parameters
        ----------
        channel_id: :class:`int`
            Discord's identifier for the channel.
        """

        if not guild_id:
            raise ValueError("Argument cannot be None: guild_id")

        roles = await self._request(Route('GET', f'/guilds/{guild_id}/roles'))

        return [Role(**role) for role in roles]

    async def get_ban(self, guild_id: int, user_id: int) -> Ban:
        """|coro|
        Get a guild ban.

        Returns
        ---------
        Optional[:class:`Ban`]
            The Ban or ``None`` if not found.

        API Documentation
        ----------
            https://discord.com/developers/docs/resources/guild#get-guild-ban

        Parameters
        ----------
        guild_id: :class:`int`
            Discord's identifier for the guild.
        user_id: :class:`int`
            Discord's identifier for the user.

        Raises
        -------
        Forbidden
            You do not have proper permissions to get the information.
        HTTPException
            An error occurred while fetching the information.
        """

        if not guild_id:
            raise ValueError("Argument cannot be None: guild_id")

        if not user_id:
            raise ValueError("Argument cannot be None: user_id")

        ban = await self._request(Route("GET", f'/guilds/{guild_id}/bans/{user_id}'))

        return Ban(**ban)

    async def get_bans(self, guild_id: int) -> List[Ban]:
        """|coro|
        Get a guild's bans.

        Returns
        ---------
        List[:class:`Role`]:
            The list of Roles.

        API Documentation
        ----------
            https://discord.com/developers/docs/resources/guild#get-guild-bans

        Parameters
        ----------
        guild_id: :class:`int`
            Discord's identifier for the guild.
        with_counts: :class:`bool`

        Raises
        -------
        Forbidden
            You do not have proper permissions to get the information.
        HTTPException
            An error occurred while fetching the information.
        """

        if not guild_id:
            raise ValueError("Argument cannot be None: guild_id")

        bans = await self._request(Route("GET", f'/guilds/{guild_id}/bans'))

        return [Ban(**bans) for ban in bans]



# -*- coding: utf-8 -*-
from .user import User

__all__ = (
    'Ban'
)

class Ban:

    """
    Model depicting a Discord ban object.
    """

    __slots__ = ('reason', 'user')

    def __init__(self, **kwargs):

        self.reason = kwargs.get('reason')
        self.user = User(**kwargs.get('user'))

    def __str__(self):
        return f'<{type(self).__name__} user_id={self.user.id}, reason={self.reason}>'



# -*- coding: utf-8 -*-
import logging
from typing import List, Optional

from aiohttp import ClientSession

from .emoji import Emoji
from .http import HTTPClient, Route

__log__ = logging.getLogger(__name__)

__all__ = (
    'EmojiClient'
)

class EmojiClient(HTTPClient):

    def __init__(self, token: str, loop=None, proxy=None, proxy_auth=None, session: Optional[ClientSession]=None) -> None:
        super().__init__(token=token, loop=loop, proxy=proxy, proxy_auth=proxy_auth, session=session)

    async def get_emoji(self, guild_id: int, emoji_id: int) -> Emoji:
        """|coro|
        Get a guild emoji.

        Returns
        ---------
        Optional[:class:`Emoji`]
            The Emoji or ``None`` if not found.

        API Documentation
        ----------
            https://discord.com/developers/docs/resources/emoji#get-guild-emoji

        Parameters
        ----------
        guild_id: :class:`int`
            Discord's identifier for the guild.
        emoji_id: :class:`int`
            Discord's identifier for the emoji.
        """

        if not guild_id:
            raise ValueError("Argument cannot be None: guild_id")

        if not emoji_id:
            raise ValueError("Argument cannot be None: emoji_id")

        emoji = await self._request(Route("GET", f'/guilds/{guild_id}/emojis/{emoji_id}'))

        return Emoji(**emoji)

    async def get_emojis(self, guild_id: int) -> List[Emoji]:
        """|coro|
        Get a guild's emojis.

        Returns
        ---------
        List[:class:`Emoji`]:
            The list of Emoji.

        API Documentation
        ----------
            https://discord.com/developers/docs/resources/emoji#list-guild-emojis

        Parameters
        ----------
        guild_id: :class:`int`
            Discord's identifier for the guild.
        """

        if not guild_id:
            raise ValueError("Argument cannot be None: guild_id")

        emojis = await self._request(Route("GET", f'/guilds/{guild_id}/emojis'))

        return [Emoji(**emojis) for emoji in emojis]



# -*- coding: utf-8 -*-
"""
The MIT License (MIT)

Copyright (c) 2015-2020 Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

class HTTPException(Exception):

    """Exception that's thrown when an HTTP request operation fails.
    Attributes
    ------------
    response: :class:`aiohttp.ClientResponse`
        The response of the failed HTTP request. This is an
        instance of :class:`aiohttp.ClientResponse`. In some cases
        this could also be a :class:`requests.Response`.
    text: :class:`str`
        The text of the error. Could be an empty string.
    status: :class:`int`
        The status code of the HTTP request.
    code: :class:`int`
        The Discord specific error code for the failure.
    """

    def __init__(self, response, message):
        self.response = response
        self.status = response.status
        if isinstance(message, dict):
            self.code = message.get('code', 0)
            base = message.get('message', '')
            errors = message.get('errors')
            if errors:
                errors = self.flatten_dict(errors)
                helpful = '\n'.join('In %s: %s' % t for t in errors.items())
                self.text = base + '\n' + helpful
            else:
                self.text = base
        else:
            self.text = message
            self.code = 0

        fmt = '{0.status} {0.reason} (error code: {1})'
        if len(self.text):
            fmt = fmt + ': {2}'

        super().__init__(fmt.format(self.response, self.code, self.text))

    def flatten_dict(self, d, key=''):
        items = []
        for k, v in d.items():
            new_key = key + '.' + k if key else k
            if isinstance(v, dict):
                try:
                    _errors = v['_errors']
                except KeyError:
                    items.extend(self.flatten_dict(v, new_key).items())
                else:
                    items.append((new_key, ' '.join(x.get('message', '') for x in _errors)))
            else:
                items.append((new_key, v))
        return dict(items)


class BadRequest(HTTPException):

    """Exception that's thrown for when status code 400 occurs."""
    pass

class Forbidden(HTTPException):

    """Exception that's thrown for when status code 403 occurs."""
    pass

class NotFound(HTTPException):

    """Exception that's thrown for when status code 404 occurs."""
    pass

class RateLimited(HTTPException):

    """Exception that's thrown for when status code 429 occurs."""

    __slots__ = ('retry_after', 'is_global')

    def __init__(self, response, message):
        super().__init__(response, message)

        self.retry_after = message['retry_after'] / 1000.0
        self.is_global = message.get('global', False)

class InternalServerError(HTTPException):

    """Exception that's thrown for when status code 500 occurs."""
    pass

class BadGateway(HTTPException):

    """Exception that's thrown for when status code 403 occurs."""
    pass


# -*- coding: utf-8 -*-
import re
from datetime import datetime

def parse_time(timestamp):
    if timestamp:
        return datetime(*map(int, re.split(r'[^\d]', timestamp.replace('+00:00', ''))))
    return None

def try_cast(value, cast_to: type):
    if not value:
        return None
    return cast_to(value)



# -*- coding: utf-8 -*-
from .snowflake import Snowflake

__all__ = (
    'PermissionOverwrite'
)

class PermissionOverwrite(Snowflake):

    __slots__ = ('type', 'allow', 'deny', 'allow_new', 'deny_new')

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.type = kwargs.get('type')
        self.allow = kwargs.get('allow')
        self.deny = kwargs.get('deny')
        self.allow_new = kwargs.get('allow_new')
        self.deny_new = kwargs.get('deny_new')



# -*- coding: utf-8 -*-
import logging
from typing import List, Optional

from aiohttp import ClientSession

from .http import HTTPClient, Route
from .voice import VoiceRegion

__log__ = logging.getLogger(__name__)

__all__ = (
    'VoiceClient'
)

class VoiceClient(HTTPClient):

    def __init__(self, token: str, loop=None, proxy=None, proxy_auth=None, session: Optional[ClientSession]=None) -> None:
        super().__init__(token=token, loop=loop, proxy=proxy, proxy_auth=proxy_auth, session=session)

    async def get_voice_regions(self) -> List[VoiceRegion]:
        """|coro|
        Get a list of voice regions.

        Returns
        ---------
        List[:class:`VoiceRegion`]:
            The list of VoiceRegions.

        API Documentation
        ----------
            https://discord.com/developers/docs/resources/voice#list-voice-regions
        """

        voice_regions = await self._request(Route('GET', '/voice/regions'))

        return [VoiceRegion(**voice_region) for voice_region in voice_regions]


import re

import setuptools

with open('requirements.txt') as f:
    REQUIREMENTS = f.readlines()

with open('README.md') as f:
    README = f.read()

with open('restcord/__init__.py') as f:
    VERSION = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE).group(1)

setuptools.setup(
    name='restcord.py',
    author='Lethys',
    url='https://github.com/Yandawl/restcord.py',
    version=VERSION,
    packages=['restcord'],
    license='MIT',
    description='An asynchronous Python client for Discord\'s API',
    long_description=README,
    long_description_content_type="text/markdown",
    keywords='discord rest api python asynchronous',
    include_package_data=True,
    install_requires=REQUIREMENTS,
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Internet',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
    ]
)



# -*- coding: utf-8 -*-
from .emoji import Emoji
from .role import Role
from .snowflake import Designation

__all__ = (
    'Guild',
    'GuildPreview'
)

class GuildPreview(Designation):

    """
    Model depicting a Discord guild object.
    """

    __slots__ = (
        'icon', 'splash', 'discovery_splash', 'features', 'emojis', 'approximate_member_count', 'approximate_presence_count', 'description'
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.icon = kwargs.get('icon')
        self.splash = kwargs.get('splash')
        self.discovery_splash = kwargs.get('discovery_splash')
        self.features = kwargs.get('features')
        self.emojis = [Emoji(**e) for e in kwargs.get('emojis', [])]
        self.approximate_member_count = kwargs.get('approximate_member_count')
        self.approximate_presence_count = kwargs.get('approximate_presence_count')
        self.description = kwargs.get('description')

class Guild(GuildPreview):

    """
    Model depicting a Discord guild object.
    """

    __slots__ = (
        'owner_id', 'application_id', 'region', 'banner', 'afk_channel_id', 'afk_timeout',
        'system_channel_id', 'widget_enabled', 'widget_channel_id', 'verification_level', 'roles', 'default_message_notifications',
        'mfa_level', 'explicit_content_filter', 'max_presences', 'max_members', 'max_video_channel_users', 'vanity_url_code', 'premium_tier', 'premium_subscription_count',
        'system_channel_flags', 'preferred_locale', 'rules_channel_id', 'public_updates_channel_id', 'embed_enabled', 'embed_channel_id'
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.owner_id = kwargs.get('owner_id')
        self.application_id = kwargs.get('application_id')
        self.region = kwargs.get('region')
        self.banner = kwargs.get('banner')
        self.afk_channel_id = kwargs.get('afk_channel_id')
        self.afk_timeout = kwargs.get('afk_timeout')
        self.system_channel_id = kwargs.get('system_channel_id')
        self.widget_enabled = kwargs.get('widget_enabled')
        self.widget_channel_id = kwargs.get('widget_channel_id')
        self.verification_level = kwargs.get('verification_level')
        self.roles = [Role(**r) for r in kwargs.get('roles', [])]
        self.default_message_notifications = kwargs.get('default_message_notifications')
        self.mfa_level = kwargs.get('mfa_level')
        self.explicit_content_filter = kwargs.get('explicit_content_filter')
        self.max_presences = kwargs.get('max_presences')
        self.max_members = kwargs.get('max_members')
        self.max_video_channel_users = kwargs.get('max_video_channel_users')
        self.vanity_url_code = kwargs.get('vanity_url_code')
        self.premium_tier = kwargs.get('premium_tier')
        self.premium_subscription_count = kwargs.get('premium_subscription_count')
        self.system_channel_flags = kwargs.get('system_channel_flags')
        self.preferred_locale = kwargs.get('preferred_locale')
        self.rules_channel_id = kwargs.get('rules_channel_id')
        self.public_updates_channel_id = kwargs.get('public_updates_channel_id')
        self.embed_enabled = kwargs.get('embed_enabled')
        self.embed_channel_id = kwargs.get('embed_channel_id')


# -*- coding: utf-8 -*-
from .snowflake import Designation

__all__ = (
    'VoiceRegion'
)

class VoiceRegion(Designation):

    """
    Model depicting a Discord voice region object.
    """

    __slots__ = ('vip', 'optimal', 'deprecated', 'custom')

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.vip = kwargs.get('vip')
        self.optimal = kwargs.get('optimal')
        self.deprecated = kwargs.get('deprecated')
        self.custom = kwargs.get('custom')



# -*- coding: utf-8 -*-
from .snowflake import Designation

__all__ = (
    'User'
)

class User(Designation):

    """
    Model depicting a Discord user object.
    """

    __slots__ = ('discriminator', 'avatar', 'public_flags')

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.name = kwargs.get('username')
        self.discriminator = kwargs.get('discriminator')
        self.avatar = kwargs.get('avatar')
        self.public_flags = kwargs.get('public_flags')

    def __str__(self):
        return f'<{type(self).__name__} id={self.id}, name={self.name}, discriminator={self.discriminator}>'

    @property
    def mention(self) -> str:
        """:class:`str`: The role's mentionable string."""
        return f'<@{self.id}>'
